import javax.swing.*;
import javax.swing.Timer;
import javax.swing.border.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.*;

/**
 * SAKExperiment - software for a text entry experiment using a scanning ambiguous keyboard (SAK).
 * <p>
 * 
 * A scanning ambiguous keyboard is a an on-screen virtual keyboard. The keyboard combines the
 * important features of scanning keyboards (input using a single physical switch or key) and
 * ambiguous keyboards (>1 letter per key). References 1 and 2 below describe the conceptual and
 * design issues as well as the initial evaluations of SAK. Reference 3, 4, and 5 describe follow-on
 * evaluations.
 * <p>
 * 
 * <h3>Related references</h3>
 * <p>
 * 
 * <ol>
 * <li>
 * <a href="http://www.yorku.ca/mack/assets2009.html"> The one-key challenge: Searching for a fast
 * one-key text entry method </a>, by MacKenzie (<i>ASSETS 2009</i>).
 * <p>
 * 
 * <li>
 * <a href="http://www.yorku.ca/mack/TOCHI2010.html"> SAK: Scanning ambiguous keyboard for efficient
 * one-key text entry </a>, by MacKenzie and Felzer (<i>TOCHI 2010</i>).
 * <p>
 * 
 * <li>
 * <a href="http://www.yorku.ca/mack/UAIS2010.html"> BlinkWrite: Efficient text entry using eye
 * blinks </a>, by MacKenzie and Ashtiani (<I>UAIS 2010</I>).
 * <p>
 * 
 * <li>
 * <a href="http://www.yorku.ca/mack/icchp2010b.html"> Qanti: A software tool for quick ambiguous
 * non-standard text input </a>, by Felzer, MacKenzie, Beckerle, and Rinderknecht (<i>ICCHP
 * 2010</i>).
 * <p>
 * 
 * <li>
 * <a href="http://www.yorku.ca/mack/ETRA2010.html"> BlinkWrite2: An improved text entry method
 * using eye blinks </a>, by Ashtiani and MacKenzie (<i>ETRA 2010</i>).
 * <p>
 * </ol>
 * <p>
 * 
 * <h3>Running the experiment software</h3>
 * <p>
 * 
 * <a href="http://www.yorku.ca/mack/HCIbook/Running/">Click here</a> for instructions on
 * launching/running the application.
 * <p>
 * 
 * <h3>Setup parameters</h3>
 * 
 * Upon launching, the application presents a setup dialog:
 * <p>
 * 
 * <center><a href="SAKExperiment-3.jpg"><img src = "SAKExperiment-3.jpg" width=400></a></center>
 * <p>
 * 
 * The default parameter settings are read from
 * <code><a href="SAKExperiment.cfg">SAKExperiment.cfg</a></code>, but may be changed through the
 * setup dialog. The setup parameters are as follows:
 * <p>
 * 
 * <ul>
 * <li>Participant code - identifies the current participant. This is used in forming the names for
 * the output data files. Also, the sd2 output data file includes a column with the participant
 * code.
 * <p>
 * 
 * <li>Condition code - an arbitrary code used to associate a test condition with this invocation.
 * This parameter might be useful if the software is used in an experiment where a condition is not
 * inherently part of the application (e.g., Gender &rarr; male, female, Input Device &rarr;
 * spacebar, headband with piezo sensor). The condition code is used in forming the name for the
 * output data file. Also, the sd2 output data file contains a column with the condition code.
 * <p>
 * 
 * <li>Session code - identifies the session. Useful if testing proceeds over multiple sessions to
 * gauge the progression of learning for <i>SAK</i>. The session code is used in forming the name
 * for the output data file. Also, the sd2 output data file contains a column with the condition
 * code.
 * <p>
 * 
 * Note: There is no setup parameter for block. The block number is generated automatically by the
 * software.
 * <p>
 * 
 * <li>Phrases per block - specifies the number of phrases presented to the participant in the
 * current block.
 * <p>
 * 
 * <li>Scan interval - specifies the scanning interval in milliseconds. The scanning interval is the
 * length of time a key is highlighted before the highlight moves to the next key.
 * <p>
 * 
 * <li>Restart scan timer on key selection - a checkbox item that, if set, causes the scan interval
 * timer to be restarted when there is a key selection. This lengthens the delay interval during the
 * current scan step. It also facilitates a novel feature of SAKs known as "multiple selection". It
 * is possible to make multiple selections in a single scan step interval if the current letter and
 * one or more subsequent letters are on the same key.
 * <p>
 * 
 * <li>Phrases file - specifies the file from which phrases are selected for input. Phrases are
 * drawn from the file at random. Typically, <code><a href="phrases2.txt">phrases2.txt</a></code> is
 * used. This is the phrase set published by MacKenzie and Soukoreff in 2003 (<a
 * href="http://www.yorku.ca/mack/chi03b.html">click here</a>).
 * <p>
 * 
 * <li>Word-freq file - specifies the word-frequency file to be used as a dictionary and for
 * encoding words and disambiguating keypresses. The default file is <code><a
 * href="d1phrases2-wordfreq.txt">d1phrases2-wordfreq.txt</a></code> which contains the core 9022
 * words and frequencies from <code>d1-wordfreq.txt</code>, which is the word-freq file used in
 * other applications in this series. The file also includes additional words that are in
 * <code>phrases2.txt</code> but not in <code>d1-wordfreq.txt</code>. This is important since SAK
 * requires that all words entered are in the dictionary.
 * <p>
 * 
 * <li>Letter assignments - specifies the letter-to-key assignments and (implicitly) the number of
 * keys. Dashes are used to separate and assign the letters to different keys. The default letter
 * assignments are abcdefgh-ijklmnop-qrstuvwxyz, divided over three keys. This results in
 * <i>KSPC</i> = 1.713 scan steps per character for English text entry. Consult references 1 and 2
 * above for all the details on why these particular letter assignments are useful.
 * <p>
 * </ul>
 * 
 * <h3>Program operation</h3>
 * 
 * After selecting OK, the program launches and presents a scanning ambiguous on-screen keyboard.
 * The keyboard has two scanning regions, one for letter selection and one for word selection.
 * Physical input uses the spacebar on the system's keyboard.
 * <p>
 * 
 * The letter-selection region organizes letters alphabetically on <i>n</i> virtual keys with more
 * than one letter per key, as per the letter-to-key assignments specified in the setup dialog. One
 * additional virtual key is added for the SPACE character.
 * <p>
 * 
 * In operation, the keys are sequentially highlighted ("scanned"). The user selects letters by
 * pressing the spacebar on the system's keyboard when the on-screen key bearing the desired letter
 * is highlighted. Selection continues in this manner until the word is spelled (or earlier, if the
 * word appears in the word list). At the end of letter selection, the user selects SPACE (by
 * pressing the spacebar when the <code>[space]</code> on-screen key is highlighted).
 * <p>
 * 
 * When the user selects SPACE, scanning enters the word-selection region. In this region, scanning
 * proceeds through a list of candidate words, ordered by their probability in English. When the
 * desired word is highlighted, the user selects it. The word is added to the transcribed text
 * followed by a SPACE character. Scanning then reverts to the letter-selection region for input of
 * the next word.
 * <p>
 * 
 * Below are two screen snaps of the application. The first shows letter selection. Scanning
 * proceeds in a repeating, left-to-right pattern.
 * <p>
 * 
 * <center><a href="SAKExperiment-1.jpg"><img src="SAKExperiment-1.jpg" width="500"></a></center>
 * <p>
 * 
 * <center><a href="SAKExperiment-2.jpg"><img src="SAKExperiment-2.jpg" width="500"></a></center>
 * <p>
 * 
 * Unlike conventional scanning keyboards, scanning does not "reset" upon selection, but, rather,
 * continues with the next key. In first the image above, the SPACE key is highlighted. The user is
 * entering "abandoned" and has correctly selected the first seven letters. The key selections
 * appear as 1112122. (The on-screen keys are numbered from 1, on the left.) At this point, the
 * desired word appears in the word selection region.
 * <p>
 * 
 * In the second screen snap, the user has selected SPACE and scanning has entered the
 * word-selection region. Scanning proceeds through the list of candidate words. As seen, the
 * desired word ("cookies") appears in the 7th position. The 4th word is currently highlighted. The
 * user must wait and then press the spacebar when the highlight reaches the desired word. Doing so
 * selects the word and adds it the transcribed text field, just below the presented text field.
 * <p>
 * 
 * 
 * If a mistake is made, a long-press on the spacebar serves as a BACKSPACE. The duration of the
 * long-press must equal or exceed the scanning interval. When a long-press is detected either the
 * current keystrokes are deleted, if entry is in the middle of a word, or the last word is deleted,
 * if entry is between words.
 * <p>
 * 
 * The application is intended for demonstrations and experimental testing. Phrases are selected at
 * random from the phrases file specified in setup. The default is <a
 * href="phrases2.txt">phrases2.txt</a>. The user enters the phrase while performance data are
 * collected. Timing begins on the first keystroke and ends when the last word in the phrase is
 * entered.
 * <p>
 * 
 * Since there is no ENTER key, the test for end-of-phrase is to compare the length of the presented
 * and transcribed text strings at the end of each word of input. When the length of the transcribed
 * text equals or exceeds the length of the presented text, an end-of-phrase is registered.
 * <p>
 * 
 * A popup window presents summary statistics at the end of each phrase. An example is shown below.
 * <p>
 * 
 * <center><a href="SAKExperiment-4.jpg"><img src="SAKExperiment-4.jpg" width="500"></a></center>
 * <p>
 * 
 * The program terminates after entry of the last phrase in a block.
 * 
 * <h3>Output data files</h3>
 * 
 * At the end of each block, data are written to two "summary data" (sd) files. The "sd1" file
 * contains data on a per-keystroke basis, one line per keystroke. The "sd2" file contains data on a
 * per-phrase basis, one line per phrase. Example data files follow:
 * <p>
 * 
 * <ul>
 * <li><a href="SAKExperiment-sd1-example.txt">sd1 example</a>
 * <li><a href="SAKExperiment-sd2-example.txt">sd2 example</a>
 * </ul>
 * 
 * The data are full-precision, comma-delimited. Importing into a spreadsheet application provides a
 * convenient method to examine the data and do follow-on analyses. <a
 * href="SAKExperiment-ExampleResults.xls">Click here</a> to open Microsoft <i>Excel</i> to view the
 * data in the example files above.
 * <p>
 * 
 * Actual output files use "SAKExperiment" as the base filename. This is followed by the participant
 * code, the condition code, the session code, and the block code, for example,
 * <code>SAKExperiment-P01-C01-S01-B01.sd1</code>.
 * <p>
 * 
 * <h3>Miscellaneous</h3>
 * 
 * When using this program in an experiment, it is a good idea to terminate all other applications
 * and disable the system's network connection. This will maintain the integrity of the data
 * collected and ensure that the program runs without hesitations.
 * <p>
 * 
 * 
 * @author Scott MacKenzie, 2008-2014
 */
public class SAKExperiment
{
	final static String APP_NAME = "SAKExperiment";
	final static int PARAMETERS = 9; // must equal the number of parameters defined below

	// setup parameters (1st value is default)
	final static String[] PARTICIPANT_CODES = { "P00", "P01", "P02", "P03", "P04", "P05", "P06", "P07", "P08", "P09",
			"P10", "P11", "P12", "P13", "P14", "P15", "P16", "P17", "P18", "P19", "P20", "P21", "P22", "P23", "P24",
			"P25", "P26", "P27", "P28", "P29", "P30" };
	final static String[] CONDITION_CODE = { "C00" };
	final static String[] SESSION_CODES = { "S00", "S01", "S02", "S03", "S04", "S05", "S06", "S07", "S08", "S09",
			"S10", "S11", "S12", "S13", "S14", "S15", "S16", "S17", "S18", "S19", "S20", "S21", "S22", "S23", "S24",
			"S25", "S26", "S27", "S28", "S29", "S30" };
	final static String[] PHRASES_PER_BLOCK = { "5" };
	final static String[] SCAN_INTERVALS = { "700", "300", "350", "400", "450", "500", "550", "600", "650", "700",
			"750", "800", "850", "900", "950", "1000", "1050", "1100", "1150", "1200", "1250", "1300", "1350", "1400",
			"1450", "1500" };
	final static String[] RESTART = { "true" };
	final static String[] PHRASES_FILE = { "phrases2.txt" };
	final static String[] WORD_FREQ_FILE = { "d1phrases2-wordfreq.txt" };
	final static String[] LETTER_ASSIGNMENTS = { "abcdefgh-ijklmnop-qrstuvwxyz" };

	// identify other files needed (used when execution from the .jar file for the 1st time)
	final static String[] OTHER_FILES = { "phrases2.txt", "d1phrases2-wordfreq.txt" };

	public static void main(String[] args) throws IOException
	{
		if (args.length != 0)
		{
			System.out.println("Usage: java SAKExperiment\n\n");
			System.out.println("Note: all configuration arguments read from SAKExperiment.cfg");
			System.exit(0);
		}

		// use Win32 look and feel
		try
		{
			UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
		} catch (Exception e)
		{
		}

		/*
		 * Define the default parameters settings. These are only used if the .cfg file does not
		 * exit, such as when the application is launched for the 1st time.
		 */
		String[] defaultParameters = new String[PARAMETERS];
		defaultParameters[0] = PARTICIPANT_CODES[0];
		defaultParameters[1] = CONDITION_CODE[0];
		defaultParameters[2] = SESSION_CODES[0];
		defaultParameters[3] = PHRASES_PER_BLOCK[0];
		defaultParameters[4] = SCAN_INTERVALS[0];
		defaultParameters[5] = RESTART[0];
		defaultParameters[6] = PHRASES_FILE[0];
		defaultParameters[7] = WORD_FREQ_FILE[0];
		defaultParameters[8] = LETTER_ASSIGNMENTS[0];

		Configuration c = Configuration.readConfigurationData(APP_NAME, defaultParameters, OTHER_FILES);
		if (c == null)
		{
			System.out.println("Error reading configuration data from SAKExperiment.cfg!");
			System.exit(0);
		}

		// ----------------------------------------------------------
		// user the app's parameter settings to create a setup dialog
		// ----------------------------------------------------------

		SetupItemInfo[] sii = new SetupItemInfo[PARAMETERS];
		sii[0] = new SetupItemInfo(SetupItem.COMBO_BOX, "Participant code ", PARTICIPANT_CODES);
		sii[1] = new SetupItemInfo(SetupItem.TEXT_FIELD, "Condition code ", CONDITION_CODE);
		sii[2] = new SetupItemInfo(SetupItem.COMBO_BOX, "Session code ", SESSION_CODES);
		sii[3] = new SetupItemInfo(SetupItem.TEXT_FIELD, "Phrases per block ", PHRASES_PER_BLOCK);
		sii[4] = new SetupItemInfo(SetupItem.COMBO_BOX, "Scan interval ", SCAN_INTERVALS);
		sii[5] = new SetupItemInfo(SetupItem.CHECK_BOX, "Restart scan timer on key selection ", RESTART);
		sii[6] = new SetupItemInfo(SetupItem.TEXT_FIELD, "Phrases file ", PHRASES_FILE);
		sii[7] = new SetupItemInfo(SetupItem.TEXT_FIELD, "Word-freq file ", WORD_FREQ_FILE);
		sii[8] = new SetupItemInfo(SetupItem.TEXT_FIELD, "Letter assignments  ", LETTER_ASSIGNMENTS);

		// use setup to allow changes to the default or existing configuration
		Setup s = new Setup(null, c, APP_NAME, sii);
		s.showSetup(null);

		SAKExperimentGUI screen = new SAKExperimentGUI(c);
		screen.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		screen.setTitle(APP_NAME);
		screen.pack();

		// put application in centre of screen
		int w = screen.getWidth();
		int h = screen.getHeight();
		Toolkit t = Toolkit.getDefaultToolkit();
		Dimension d = t.getScreenSize();
		screen.setLocation((d.width - w) / 2, (d.height - h) / 2);
		screen.setVisible(true);
	}
}

// -----------------------------
// define the application window
// -----------------------------

class SAKExperimentGUI extends JFrame implements ActionListener, KeyListener
{
	// the following avoids a "warning" with Java 1.5.0 complier (?)
	static final long serialVersionUID = 42L;

	// -----------------------------------------------------------------
	// declare variables and components accessed by more than one method
	// -----------------------------------------------------------------

	final Font NORMAL_FONT = new Font("sanserif", Font.PLAIN, 18);
	final Font HIGHLIGHT_FONT = new Font("sanserif", Font.BOLD, 18);
	final Font KEY_FONT = new Font("sanserif", Font.PLAIN, 24);
	final Font KEY_BOLD_FONT = new Font("sanserif", Font.BOLD, 24);
	final int QUERY_MAX = 50;
	final String SD1_HEADER = "t,mode,key_num,key_code,selected,number_of_selections\n";
	final String SD2_HEADER = "App,Participant,Condition,Session,Block,Scan_interval,Letter-key_assignment,"
			+ "Phrase,Presented_text,Min_scans,Transcribed_text," + "Time,Scans,Selections,WPM,ER,Efficiency,"
			+ "Multiple_select_count,Multiple_select_opportunities,"
			+ "Early_word_select,Early_word_select_opportunities," + "Phrase_retries,Word_deletes\n";
	Timer t;
	int activeKey;
	int activeOptionKey;
	Color back;
	final Color ACTIVE_COLOR = Color.PINK;
	JTextField presentedTextField;
	JTextField transcribedTextField;
	JTextField currentKeysField;
	JLabel[] queryField;
	String transcribedText;
	String presentedText;
	String currentKeys;
	JPanel keyboardPanel;
	String letters;
	JPanel queryPanel;
	int mode;
	final int LETTER = 0;
	final int WORD = 1;
	final String[] MODE = { "letter", "word" };
	int idx = 0;
	long elapsedTime;
	JDialog resultsDialog;
	JOptionPane resultsPane;
	JTextArea resultsArea;
	String ALPHABETIC = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
	int letterKeys;
	String[] keyLabel;
	JButton[] key; // = new JButton[letterKeys + 1];
	JButton[] optionKey = new JButton[3];
	EncodedWord[] ew;
	String[] query;
	String[] phrase;
	long timeStamp;
	boolean begin; // false until first "keypress" in a phrase
	boolean event; // true if a key is pressed during a scanning interval
	String keyCode;
	int eventMode;
	int scanCount;
	int minScanCount;
	int scanInterval;
	boolean timerRestart;
	String phrasesFile;
	String wordFreqFile;
	String scanString;
	BufferedWriter sd1, sd2;
	int phrasesPerBlock;
	int blockNumber;
	int selections;
	String participantCode, conditionCode, sessionCode, blockCode;
	String letterKeyAssignment;

	// -----------
	// constructor
	// -----------

	public SAKExperimentGUI(Configuration c) throws IOException
	{
		// --------------------------------------
		// declare and initialize local variables
		// --------------------------------------

		participantCode = c.getConfigurationParameter(0);
		conditionCode = c.getConfigurationParameter(1);
		sessionCode = c.getConfigurationParameter(2);
		// NOTE: block code computed automatically
		phrasesPerBlock = Integer.parseInt(c.getConfigurationParameter(3));
		scanInterval = Integer.parseInt(c.getConfigurationParameter(4));
		timerRestart = Boolean.valueOf(c.getConfigurationParameter(5));
		phrasesFile = c.getConfigurationParameter(6);
		wordFreqFile = c.getConfigurationParameter(7);
		letterKeyAssignment = c.getConfigurationParameter(8);

		scanString = "";
		scanCount = 0;
		minScanCount = 0;

		String[] tmp = letterKeyAssignment.split("-");
		letterKeys = tmp.length;
		keyLabel = new String[letterKeys + 1];
		key = new JButton[letterKeys + 1];
		for (int i = 0; i < keyLabel.length - 1; ++i)
			keyLabel[i] = tmp[i].toUpperCase();
		keyLabel[tmp.length] = "[space]";

		t = new Timer(scanInterval, this);
		t.start();
		activeKey = 0;
		mode = LETTER;
		timeStamp = 0;
		elapsedTime = 0;
		begin = false;
		event = false;
		letters = ALPHABETIC;
		transcribedText = "";
		currentKeys = "";
		selections = 0;

		// -----------------------
		// read phrases into array
		// -----------------------

		BufferedReader br = null;
		try
		{
			br = new BufferedReader(new FileReader(phrasesFile));
		} catch (FileNotFoundException e)
		{
			System.out.println("File not found: " + phrasesFile);
			System.exit(0);
		}
		String line = null;
		Vector<String> v = new Vector<String>();
		while ((line = br.readLine()) != null)
			v.addElement(br.readLine());
		phrase = new String[v.size()];
		v.copyInto(phrase);

		// ------------------
		// configure keyboard
		// ------------------

		StringTokenizer st = new StringTokenizer(letterKeyAssignment, "-");
		String[] keyboard = new String[st.countTokens()];
		for (int i = 0; i < keyboard.length; ++i)
			keyboard[i] = st.nextToken();

		// --------------------------------------------
		// read word-freq file, build EncodedWord array
		// --------------------------------------------

		br = null;
		try
		{
			br = new BufferedReader(new FileReader(wordFreqFile));
		} catch (FileNotFoundException e)
		{
			System.out.println("File not found: " + wordFreqFile);
			System.exit(0);
		}
		Vector<String> w = new Vector<String>();
		Vector<String> f = new Vector<String>();
		while ((line = br.readLine()) != null)
		{
			st = new StringTokenizer(line);
			if (st.countTokens() != 2)
			{
				System.out.println("Format error in word-freq file!");
				System.exit(0);
			}
			w.addElement(st.nextToken());
			f.addElement(st.nextToken());
		}
		String[] wArray = new String[w.size()];
		String[] fArray = new String[f.size()];
		w.copyInto(wArray);
		f.copyInto(fArray);
		ew = new EncodedWord[wArray.length];
		for (int i = 0; i < ew.length; ++i)
		{
			String word = wArray[i].toLowerCase();
			int freq = Integer.parseInt(fArray[i]);
			String keystrokes = "";
			for (int j = 0; j < word.length(); ++j)
			{
				for (int k = 0; k < keyboard.length; ++k)
				{
					if (keyboard[k].indexOf(word.substring(j, j + 1)) >= 0)
					{
						keystrokes += k + 1;
						break;
					}
				}
			}
			ew[i] = new EncodedWord(word, freq, keystrokes);
		}
		Arrays.sort(ew, new ByCode());

		// -------------------
		// open sd1, sd2 files
		// -------------------

		blockCode = "";
		int blockCodeNumber = 0;
		String base = "SAKExperiment-" + participantCode + "-" + conditionCode + "-" + sessionCode;
		String s1 = "";
		String s2 = "";
		do // find next available block code
		{
			++blockCodeNumber;
			blockCode = blockCodeNumber < 10 ? "B0" + blockCodeNumber : "B" + blockCodeNumber;
			s1 = base + "-" + blockCode + ".sd1";
			s2 = base + "-" + blockCode + ".sd2";
		} while (new File(s1).exists());

		try
		{
			sd1 = new BufferedWriter(new FileWriter(s1));
			sd2 = new BufferedWriter(new FileWriter(s2));
		} catch (IOException e)
		{
			System.out.println("I/O error: can't open sd1/sd2 data files");
			System.exit(0);
		}

		// initialize array of keys
		for (int i = 0; i < letterKeys + 1; ++i)
		{
			key[i] = new JButton(keyLabel[i]);
			key[i].addKeyListener(this);
			key[i].setFont(KEY_FONT);
		}
		key[0].setFont(KEY_BOLD_FONT);

		// write header lines to sd1 and sd2 files
		writeData(sd1, SD1_HEADER);
		writeData(sd2, SD2_HEADER);

		// -------------------------------
		// create and configure components
		// -------------------------------

		back = (new Button()).getBackground();
		presentedTextField = new JTextField(50);
		presentedText = getNextPhrase(phrase);
		presentedTextField.setText(presentedText);
		presentedTextField.setFont(NORMAL_FONT);
		presentedTextField.setBackground(back);
		presentedTextField.setFocusable(false);
		presentedTextField.setBorder(new TitledBorder(new EtchedBorder(), "Presented Text"));

		transcribedTextField = new JTextField(50);
		transcribedTextField.setFont(NORMAL_FONT);
		transcribedTextField.setBackground(back);
		transcribedTextField.setFocusable(false);
		transcribedTextField.setBorder(new TitledBorder(new EtchedBorder(), "Transcribed Text"));

		currentKeysField = new JTextField(50);
		currentKeysField.setFont(NORMAL_FONT);
		currentKeysField.setBackground(back);
		currentKeysField.setFocusable(false);
		currentKeysField.setBorder(new TitledBorder(new EtchedBorder(), "Currenty Keys"));

		queryField = new JLabel[QUERY_MAX];
		queryPanel = new JPanel();
		queryPanel.setLayout(new FlowLayout(FlowLayout.LEFT, 10, 1));
		for (int i = 0; i < QUERY_MAX; ++i)
		{
			queryField[i] = new JLabel();
			queryField[i].setFont(NORMAL_FONT);
			queryField[i].setBorder(new LineBorder(this.getBackground(), 1));
			queryField[i].setBackground(back);
			queryField[i].setFocusable(false);
			queryField[i].setText(" ");
			queryPanel.add(queryField[i]);
			queryPanel.setPreferredSize(new Dimension(900, 150));
			queryPanel.setBorder(new TitledBorder(new EtchedBorder(), "Candidate Words"));
		}

		keyboardPanel = new JPanel();
		keyboardPanel.setPreferredSize(new Dimension(900, 100));
		keyboardPanel.setLayout(new GridLayout(1, letterKeys + 1, 10, 10));
		for (int i = 0; i < letterKeys + 1; ++i)
			keyboardPanel.add(key[i]);

		resultsArea = new JTextArea(9, 30);
		resultsArea.setFont(new Font("sansserif", Font.PLAIN, 18));
		resultsArea.setBackground(back);
		resultsPane = new JOptionPane(resultsArea, JOptionPane.INFORMATION_MESSAGE);
		resultsPane.setFont(new Font("sansserif", Font.PLAIN, 18));
		resultsDialog = resultsPane.createDialog(this, "Information");

		// ------------------
		// arrange components
		// ------------------

		JPanel northPanel = new JPanel();
		northPanel.setBorder(BorderFactory.createEmptyBorder(30, 30, 30, 30));
		northPanel.setLayout(new BoxLayout(northPanel, BoxLayout.Y_AXIS));
		northPanel.add(Box.createRigidArea(new Dimension(0, 10)));
		northPanel.add(presentedTextField);
		northPanel.add(Box.createRigidArea(new Dimension(0, 10)));
		northPanel.add(transcribedTextField);
		northPanel.add(Box.createRigidArea(new Dimension(0, 30)));
		northPanel.add(keyboardPanel);
		northPanel.add(Box.createRigidArea(new Dimension(0, 30)));
		northPanel.add(currentKeysField);
		northPanel.add(Box.createRigidArea(new Dimension(0, 30)));
		northPanel.add(queryPanel);
		northPanel.add(Box.createRigidArea(new Dimension(0, 30)));
		this.setContentPane(northPanel);
	}

	private String getNextPhrase(String[] phraseArg)
	{
		Random r = new Random();
		int n = r.nextInt(phraseArg.length);
		return phraseArg[n];
	}

	// -------------------------------
	// implement ActionListener method
	// -------------------------------

	String s = "";
	int idxSave = 0;
	int keyNumSave = 0;
	int phraseCount = 0;
	int phraseRetry = 0;
	int wordDeletion = 0;
	boolean startCandidate = false;

	public void actionPerformed(ActionEvent ae)
	{
		if (keyTypedCount > 0)
		{
			t.stop();
			if (currentKeys.length() > 1) // always at least one key at this juncture
			{
				writeData(sd1, "[restart_current_word]\n");
				++wordDeletion;
				currentKeys = "";
				currentKeysField.setText("");
				clearQueryPanel();
			} else if (transcribedText.length() > 2)
			{
				if (currentKeys.length() == 0)
					writeData(sd1, "[abandon_word_select]\n");
				else
					writeData(sd1, "[delete_last_word]\n");

				++wordDeletion;
				currentKeys = "";
				currentKeysField.setText("");
				transcribedText = deleteLastWord(transcribedText);
				transcribedTextField.setText(transcribedText);
				clearQueryPanel();
			}
			mode = LETTER; // stay in (return to!) letter mode
		}

		++scanCount;
		elapsedTime = System.currentTimeMillis() - timeStamp;
		s = "" + elapsedTime;

		key[activeKey].setFont(KEY_FONT);
		if (mode == LETTER)
		{
			// check if there was double selection
			if (checkMultiple == 2)
				++multipleSelectCount;

			if (event) // SPACE was selected in the last scan interval (via KeyEvent)
			{
				s += "," + MODE[eventMode] + "," + keyNumSave + "," + keyCode;
				if (activeKey == key.length - 1)
					scanString += "W";
				else
					scanString += "" + (activeKey + 1); // key #
			} else
			// a key was *not* pressed in the last scan interval
			{
				// info on last key
				s += "," + MODE[mode] + "," + activeKey + "," + key[activeKey].getText();
				scanString += "."; // pass
			}
			++activeKey;
			if (activeKey == key.length)
				activeKey = 0;
			key[activeKey].setFont(KEY_BOLD_FONT);
			key[activeKey].requestFocus();
		} else if (mode == WORD)
		{
			if (event)
			{
				s += "," + MODE[eventMode] + "," + activeKey + "," + keyCode;
				scanString += "S"; // select candidate word (last scan period)
			} else
			{
				if (query != null)
				{
					int tmp99 = (idx - 1) < 0 ? 0 : (idx - 1);
					// info on last key
					s += "," + MODE[mode] + "," + (tmp99) + "," + query[tmp99];
				} else
					s += "," + MODE[mode] + "," + "?";
				scanString += ".";
			}
			idx = updateQueryPanel(query, idx);
			++idx;
		}
		s += "," + event + "," + checkMultiple + "\n";
		if (begin)
			writeData(sd1, s);
		event = false;
		checkMultiple = 0;
	}

	// -----------------------------
	// implement KeyListener methods
	// -----------------------------

	int keyTypedCount = 0;

	public void keyPressed(KeyEvent ke)
	{
	}

	public void keyReleased(KeyEvent ke)
	{
		keyTypedCount = 0;
		t.start();
	}

	int checkMultiple = 0;
	int multipleSelectCount = 0; // count number of MLPK opportunities taken
	int earlyWordSelect = 0; // count number of early word selections

	public void keyTyped(KeyEvent ke)
	{
		++keyTypedCount;
		if (keyTypedCount > 1)
			return;
		else if (timerRestart)
			t.restart();
		++selections;
		++checkMultiple;

		// Hit 'q' to abort current phrase and start a new phrase
		if (("" + ke.getKeyChar()).equals("q"))
		{
			Toolkit.getDefaultToolkit().beep();
			writeData(sd1, "[abort]\n");
			++phraseRetry;
			--phraseCount;
			presentedText = getNextPhrase(phrase);
			presentedTextField.setText(presentedText);
			minScanCount = getMinimumScans(presentedText).length();
			transcribedText = "";
			transcribedTextField.setText(transcribedText);
			currentKeys = "";
			currentKeysField.setText(currentKeys);
			mode = LETTER;
			idx = 0;
			clearQueryPanel();
			begin = false;
			return;
		}
		event = true;
		eventMode = mode;
		if (!begin)
		{
			++phraseCount;
			writeData(sd1, "#\n" + presentedText + "\n");
			begin = true;
			minScanCount = getMinimumScans(presentedText).length();
			scanCount = 0;
			timeStamp = System.currentTimeMillis();
			scanString = "";
		}
		if (mode == LETTER)
		{
			keyCode = key[activeKey].getText();
			keyNumSave = activeKey;
			if (activeKey == key.length - 1) // virtual [space] key
			{
				startCandidate = true;

				if (query != null)
					mode = WORD;
				else
				{
					Toolkit.getDefaultToolkit().beep();
				}
				return;
			}
			currentKeys += "" + (activeKey + 1);
			currentKeysField.setText(currentKeys);
			query = getCandidates(currentKeys, ew);
			updateQueryPanel(query, -1);
			if (query == null)
			{
				Toolkit.getDefaultToolkit().beep();
				currentKeys = "";
				currentKeysField.setText("");
				clearQueryPanel();
			}
		} else if (mode == WORD) // a candidate word has been selected
		{
			if (query != null)
			{
				idx = idx > 0 ? idx - 1 : 0;
				keyCode = query[idx];
				keyNumSave = idx;
				idxSave = idx;
				transcribedText += query[idx] + " ";
				String lastWord = query[idx];
				query = null;
				transcribedTextField.setText(transcribedText);

				if (currentKeys.length() < lastWord.length())
					++earlyWordSelect;

				/*
				 * Test to see if the phrase is finished being entered. Since there is no ENTER key,
				 * the test is based on a comparison between the length of the transcribed and
				 * presented text. Obviously, this is not ideal, since dropping a word or a few
				 * characters would leave the transcribed text short and thus leave the application
				 * waiting for more input.
				 */
				if (transcribedText.length() >= presentedText.length())
				{
					// final write needed here
					elapsedTime = System.currentTimeMillis() - timeStamp;
					s = elapsedTime + "," + MODE[mode] + "," + idx + "," + lastWord + "," + event + ",1" + "\n";
					writeData(sd1, s);

					++scanCount; // inc now because final timeout missed trim trailing space
					transcribedText = transcribedText.substring(0, transcribedText.length() - 1);

					Toolkit.getDefaultToolkit().beep();
					begin = false;
					scanString += "W"; // final adjustment needed
					String minScanString = getMinimumScans(presentedText);
					double efficiency = ((double)minScanString.length() / scanString.length()) * 100.0;

					int letterCount = transcribedText.length();
					double wpm = (letterCount / 5.0) / (elapsedTime / 60000.0);
					String s1 = presentedText.toLowerCase();
					String s2 = transcribedText.toLowerCase();
					double er = (new MSD(s1, s2)).getErrorRateNew();

					writeData(sd2, "SAKExperiment," + participantCode + "," + conditionCode + "," + sessionCode + ","
							+ blockCode + "," + scanInterval + "," + letterKeyAssignment + "," + phraseCount + ","
							+ presentedText.toLowerCase() + "," + minScanString.length() + ","
							+ transcribedText.toLowerCase() + "," + elapsedTime + "," + scanString.length() + ","
							+ selections + "," + wpm + "," + er + "," + efficiency + "," + multipleSelectCount + ","
							+ getMultipleSelectOpportunities(presentedText) + "," + earlyWordSelect + ","
							+ getEarlyWordSelectOpportunities(presentedText) + "," + phraseRetry + "," + wordDeletion
							+ "\n");

					keyTypedCount = 0; // seem to need this before the popup

					// build formated results to present in GUI
					String s = "Thank you.  Press SPACE to continue.\n\n";
					s += String.format("Entry speed = %.2f wpm\n", wpm);
					s += String.format("Error rate = %.2f%%\n", er);
					s += String.format("Scan step count = %d\n", scanCount);
					s += String.format("Minimum scan steps = %d\n", minScanCount);
					s += String.format("Efficiency = %.2f%%\n", efficiency);
					resultsArea.setText(s);
					resultsDialog.setVisible(true);

					// check to see if we're done
					if (phraseCount == phrasesPerBlock)
					{
						JLabel thanks = new JLabel("End of Block. Thank you!");
						thanks.setFont(new Font("sanserif", Font.PLAIN, 18));
						JOptionPane.showMessageDialog(this, thanks, "Message", JOptionPane.INFORMATION_MESSAGE);
						System.exit(0);
					}

					// prepare for next phrase
					earlyWordSelect = 0;
					multipleSelectCount = 0;
					selections = 0;
					t.start();
					phraseRetry = 0;
					wordDeletion = 0;
					presentedText = getNextPhrase(phrase);
					presentedTextField.setText(presentedText);
					minScanCount = getMinimumScans(presentedText).length();
					transcribedText = "";
					transcribedTextField.setText(transcribedText);
					currentKeys = "";
					currentKeysField.setText(currentKeys);
					mode = LETTER;
					idx = 0;
					clearQueryPanel();
					return;
				}
			} else
			{
				Toolkit.getDefaultToolkit().beep();
			}
			currentKeys = "";
			currentKeysField.setText(currentKeys);
			mode = LETTER;
			idx = 0;
			clearQueryPanel();
			return;
		}
	}

	private String deleteLastWord(String sArg)
	{
		String s = sArg;
		s = s.trim();
		if (s.length() < 2)
			return s;
		int foundAt = -1;
		for (int i = s.length() - 1; i > 0; --i)
			if (s.charAt(i) == ' ')
			{
				foundAt = i;
				break;
			}
		if (foundAt > 0)
			return s.substring(0, foundAt + 1); // leave trailing space (for now)
		else
			return "";
	}

	private void writeData(BufferedWriter bw, String s)
	{
		try
		{
			bw.write(s, 0, s.length());
			bw.flush();
		} catch (IOException e)
		{
			System.out.println("Error writing to sd1 file!");
			System.exit(1);
		}
	}

	private String getMinimumScans(String phraseArg)
	{
		String minScans = "";
		StringTokenizer st = new StringTokenizer(phraseArg);
		while (st.hasMoreTokens())
			minScans += SAKStats.getScanSequence(SAKStats.OW, st.nextToken(), ew);
		int i = 0;
		while (minScans.charAt(i) == '.')
			minScans = minScans.substring(1); // remove scans before 1st selection
		return minScans;
	}

	private int getMultipleSelectOpportunities(String phraseArg)
	{
		int count = 0;
		StringTokenizer st = new StringTokenizer(phraseArg);
		while (st.hasMoreTokens())
			count += SAKStats.getMLPKOpportunities(st.nextToken(), ew);
		return count;
	}

	private int getEarlyWordSelectOpportunities(String phraseArg)
	{
		int count = 0;
		StringTokenizer st = new StringTokenizer(phraseArg.toLowerCase());
		while (st.hasMoreTokens())
			count += SAKStats.getOWOpportunities(st.nextToken(), ew);
		return count;
	}

	private String[] getCandidates(String keycodeArg, EncodedWord[] ewArg)
	{
		final int MAX = 15; // maximum reasonable list size (before full word entered)
		String keycode = keycodeArg;
		EncodedWord[] ew = ewArg;

		String[] unique = EncodedWord.getUnique(keycode, ew);
		String[] query = EncodedWord.getQuery(keycode, ew);

		// if nothing happening, return null
		int n1 = unique == null ? 0 : unique.length;
		int n2 = query == null ? 0 : query.length;
		if ((n1 + n2) == 0)
			return null;

		String[] candidates;
		if (n1 >= MAX)
		{
			candidates = new String[n1];
			for (int i = 0; i < n1; ++i)
				candidates[i] = unique[i];
		} else
		{
			int total = n2 > MAX ? MAX : n2;
			candidates = new String[total];
			for (int i = 0; i < total; ++i)
				candidates[i] = query[i];
		}
		return candidates;
	}

	private int updateQueryPanel(String[] queryArg, int highlightArg)
	{
		String[] query = queryArg;
		int highlight = highlightArg;
		if (query == null)
			return -1;
		if (highlight >= QUERY_MAX || highlight >= query.length)
			highlight = 0;
		clearQueryPanel();

		// adjust limit if number of candidates exceeds QUERY_MAX
		int numberOfQueries = query.length > QUERY_MAX ? QUERY_MAX : query.length;
		for (int i = 0; i < numberOfQueries; ++i)
		{
			queryField[i].setText(" " + query[i] + " ");
			if (highlight > 0)
			{
				queryField[highlight - 1].setForeground(Color.BLACK);
				queryField[highlight - 1].setBorder(new LineBorder(this.getBackground(), 1));
			}
			if (highlight >= 0)
			{
				queryField[highlight].setForeground(Color.BLUE);
				queryField[highlight].setBorder(new LineBorder(Color.GRAY, 1));
			}
			if (i == QUERY_MAX - 1)
				break;
		}
		if (highlight < QUERY_MAX - 1)
			return highlight;
		else
			return -1;
	}

	private void clearQueryPanel()
	{
		for (int i = 0; i < queryField.length; ++i)
		{
			queryField[i].setText(" ");
			queryField[i].setForeground(Color.BLACK);
			queryField[i].setBorder(new LineBorder(this.getBackground(), 1));
		}
		return;
	}
}