import java.io.*;
import java.util.*;

/** SAKStats - a program to provide stats and other services related to the 
* scanning ambiguous keyboard (SAK) implemented in OneKey.java.
* <p>
*
* Invocation:<p>
*
* <pre>
*     PROMPT>java SAKStats file letters [-olps|-mlps|-mlpk|-ow] [-wfk|-wfks|-spc|-sps]
*
*     where  file = word-freq dictionary file
*            letters = letter-key assignment (e.g., abcdefg-hijklmn-opqrstu-vwxyz)
*            -olps = one-letter-per-scan method
*            -mlps = multiple-letter-per-scan method
*            -mlpk = multiple-letter-per-key method
*            -ow   = optimized-word method
*            -wfk  = output word-freq-keycode data
*            -wfks = output word-freq-keycode-scan data
*            -spc  = output scans per character stat
*            -sps  = output selections per scan stat
* </pre>
*
* Note: "mlps", above, means multiple letter selections during a single scan through
* the letter-selection region.  "mlpk" means multiple letter selections during
* a single scan step interval where multiple desired letters are on the same key.
* <p>
*
* Here are some sample runs:<p>
* 
* <pre>
*     PROMPT>java SAKStats d1phrases2-wordfreq.txt abcdefgh-ijklmnop-qrstuvwxyz -ow -spc
*     abcdefgh-ijklmnop-qrstuvwxyz ow 1.7131456157916434
*
*     PROMPT>java SAKStats d1phrases2.txt abcdefgh-ijklmnop-qrstuvwxyz -ow -sps
*     abcdefgh-ijklmnop-qrstuvwxyz ow 0.603138251036028
* </pre>
*
* Here are a few runs generating word-freq-keycode-scan data:<p>
*
* <pre>
*     PROMPT>java SAKStats d1phrases2-wordfreq.txt abcdefghijklm-nopqrstuvwxyz -wfks -ow > onekey-k2ks-scans.txt
*     PROMPT>java SAKStats d1phrases2-wordfreq.txt abcdefgh-ijklmnop-qrstuvwxyz -wfks -ow > onekey-k3ks-scans.txt
*     PROMPT>java SAKStats d1phrases2-wordfreq.txt abcdefgh-ijklm-nopqr-stuvwxyz -wfks -ow > onekey-k4ks-scans.txt
*     PROMPT>java SAKStats d1phrases2-wordfreq.txt abcdefgh-ijklm-nopqr-stu-vwxyz -wfks -ow > onekey-k5ks-scans.txt
*     PROMPT>java SAKStats d1phrases2-wordfreq.txt abcdefgh-ijkl-mno-pqr-stu-vwxyz -wfks -ow > onekey-k6ks-scans.txt
* </pre>
*
* To view the OW output files, click below.
*
* <ul>
* <li><a href="onekey-k2ks-scans.txt">onekey-k2ks-scans.txt</a>
* <li><a href="onekey-k3ks-scans.txt">onekey-k3ks-scans.txt</a>
* <li><a href="onekey-k4ks-scans.txt">onekey-k4ks-scans.txt</a>
* <li><a href="onekey-k5ks-scans.txt">onekey-k5ks-scans.txt</a>
* <li><a href="onekey-k6ks-scans.txt">onekey-k6ks-scans.txt</a>
* </ul>
* 
* @see <a href="SAKStats.java">source code</a>
* @author Scott MacKenzie, 2008-2011
*/

public class SAKStats
{
	final static int OLPS = 0;
	final static int MLPS = 1;
	final static int MLPK = 2;
	final static int OW = 3;

	public static void main(String[] args) throws IOException
	{
		if (args.length == 0 || args.length < 3)
			usage();

		String fileName = args[0];
		StringTokenizer st = new StringTokenizer(args[1], "-");
		String[] keyboard = new String[st.countTokens()];
		for (int i = 0; i < keyboard.length; ++i)
			keyboard[i] = st.nextToken();

		// method options (each method encompasses the previous method)
		boolean olpsOption = false; // one-letter-per-scan method 
		boolean mlpsOption = false; // multiple-letter-per-scan method
		boolean mlpkOption = false; // multiple-letter-per-key method
		boolean owOption = false;   // optimized-word method

		// output options
		boolean wfkOption = false;  // word-frequency-keycode
		boolean wfksOption = false; // word-frequency-keycode-scan 
		boolean spcOption = false;  // scans per character
		boolean spsOption = false;  // selections per scan

		for (int i = 2; i < args.length; ++i)
		{
			if (args[i].equals("-olps"))  olpsOption = true;
			if (args[i].equals("-mlps"))  mlpsOption = true;
			if (args[i].equals("-mlpk"))  mlpkOption = true;
			if (args[i].equals("-ow"))    owOption = true;
			if (args[i].equals("-wfk"))   wfkOption = true;
			if (args[i].equals("-wfks"))  wfksOption = true;
			if (args[i].equals("-spc"))   spcOption = true;
			if (args[i].equals("-sps"))   spsOption = true;
		}

		int mode;
		if (olpsOption) mode = OLPS;
		else if (mlpsOption) mode = MLPS;
		else if (mlpkOption) mode = MLPK;
		else mode = OW;

		if (!olpsOption && !mlpsOption && !mlpkOption && !owOption)
		{
			System.out.println("*** Error: must specify a method ***");
			usage();
		}

		EncodedWord[] ew = buildEncodedWordArray(fileName, keyboard);

		String word = "";
		int freq = 0;
		String keystrokes = "";
		String scanString = "";
		long totalScans = 0;
		long totalSelections = 0;
		long totalCharacters = 0;
		int selectionCount = 0;

		for (int i = 0; i < ew.length; ++i)
		{
			word = ew[i].getWord();
			freq = ew[i].getFreq();
			keystrokes = ew[i].getCode();
			scanString = getScanSequence(mode, word, ew);
			selectionCount = getSelectionCount(mode, word, keystrokes, ew);

			if (wfksOption)
				System.out.println(word + " " + freq + " " + keystrokes + " " + scanString);
			else if (wfkOption)
				System.out.println(word + " " + freq + " " + keystrokes);

			totalCharacters += (word.length() + 1) * freq; // +1 for SPACE
			totalScans += scanString.length() * freq;
			totalSelections += selectionCount * freq;
		}

		// compute/output scans per character (SPC), and selections per scan (SPS)
		double spc = 0.0;
		double sps = 0.0;
		double wpmExample = 0.0;
		spc = (double)totalScans / totalCharacters;
		sps = (double)totalSelections / totalScans;
		wpmExample = 700.0 * spc;           // ms per char
		wpmExample = wpmExample / 60000.0;  // minutes per char
		wpmExample = 1.0 / wpmExample;      // char per minute
		wpmExample = wpmExample / 5.0;      // wpm

		if (spcOption || spsOption)
		{
			String k = "";
			for (int i = 0; i < keyboard.length; ++i)
			{
				k += keyboard[i];
				if (i < keyboard.length - 1)
					k += "-";
			}
			if (spcOption)
				System.out.println("" + k + " " + args[2].substring(1) + " " + spc);
			else if (spsOption)
				System.out.println("" + k + " " + args[2].substring(1) + " " + sps);
		}
	}

	// ----------------------------------------------------------------------
	// return a scan sequence for the specified word using the specified mode
	// ----------------------------------------------------------------------
	public static String getScanSequence(int modeArg, 
		String wordArg, EncodedWord[] ewArg)
	{
		if (modeArg == SAKStats.OLPS)
			return getOLPSScanSequence(wordArg.toLowerCase(), ewArg);
		else if (modeArg == SAKStats.MLPS)
			return getMLPSScanSequence(wordArg.toLowerCase(), ewArg);
		else if (modeArg == SAKStats.MLPK)
			return getMLPKScanSequence(wordArg.toLowerCase(), ewArg);
		else if (modeArg == SAKStats.OW)
			return getOWScanSequence(wordArg.toLowerCase(), ewArg);			
		else
			System.out.println("getScanSequence error: no or invalid mode argument!");
		return "?";
	}

	//-------------------------------------------------------------------------------
	// OLPS = One Letter Per Scan scanning method
	private static String getOLPSScanSequence(String wordArg, EncodedWord[] ewArg)
	{
		String word = wordArg;
		EncodedWord[] ew = ewArg;
		int[] letterPosition = getLetterPosition(ew);
		String[] keyboard = getKeyboard(letterPosition);
		String keystrokes = getKeystrokes(word, letterPosition);
		String scanString = "";
		for (int j = 0; j < word.length(); ++j)
		{
			char c = word.charAt(j);
			int p = letterPosition[c - 'a'];
			for (int k = 0; k < p; ++k)
				scanString += ".";
			scanString += c;
			for (int k = 0; k < keyboard.length - p; ++k)
				scanString += ".";
		}
		scanString = scanString.substring(0, scanString.length() - 1); // drop last "."
		scanString += "S";
		String[] query = getCandidates(keystrokes, ew);
		if (query == null)
			return null;
		for (int k = 0; k < query.length; ++k)
			if (word.equals(query[k]))
				break;
			else
				scanString += ".";
		scanString += "W";
		return scanString;
	}

	//-------------------------------------------------------------------------------
	// MLPS = Multiple Letter Per Scan scanning method
	private static String getMLPSScanSequence(String wordArg, EncodedWord[] ewArg)
	{
		String word = wordArg;
		EncodedWord[] ew = ewArg;
		int[] letterPosition = getLetterPosition(ew);
		String[] keyboard = getKeyboard(letterPosition);
		String keystrokes = getKeystrokes(word, letterPosition);
		int dist[][] = getDistanceMatrix(letterPosition);
		String scanString = "";

		char c1 = word.charAt(0);
		char c2;
		scanString = "";
		for (int j = 0; j < letterPosition[c1 - 'a']; ++j)
			scanString += ".";
		scanString += c1;		

		for (int j = 1; j < word.length(); ++j)
		{
			c1 = word.charAt(j - 1);
			c2 = word.charAt(j);
			int idx1 = c1 - 'a';
			int idx2 = c2 - 'a';
			int d = dist[idx1][idx2];
			for (int k = 0; k < d - 1; ++k)
				scanString += ".";
			scanString += c2;
		}
		c2 = word.charAt(word.length() - 1); // last letter in word
		int n = letterPosition[c2 - 'a'];
		for (int k = 0; k < keyboard.length - n - 1; ++k)
			scanString += ".";
		scanString += "S";
		String[] query = getCandidates(keystrokes, ew);
		if (query == null)
			return null;
		for (int k = 0; k < query.length; ++k)
			if (word.equals(query[k]))
				break;
			else
				scanString += ".";
		scanString += "W";
		return scanString;
	}

	//-------------------------------------------------------------------------------
	// MLPK = Multiple Letter Per Key scanning method
	private static String getMLPKScanSequence(String wordArg, EncodedWord[] ewArg)
	{
		String word = wordArg;
		EncodedWord[] ew = ewArg;
		int[] letterPosition = getLetterPosition(ew);
		String[] keyboard = getKeyboard(letterPosition);
		String keystrokes = getKeystrokes(word, letterPosition);
		int dist[][] = getDistanceMatrix(letterPosition);
		String scanString = "";

		String reducedWord = "";
		char c1, c2;

		// Reduce the word!  Where multiple letters are on the same key, 
		// only show the 1st letter of the group	
		for (int j = 0; j < keystrokes.length(); ++j)
		{
			reducedWord += word.charAt(j);
			while (j < keystrokes.length() - 1 && 
					keystrokes.charAt(j) == keystrokes.charAt(j + 1)) 
				++j;
		}		

		// now proceed with "reduced word", as per MLPS method
		c1 = reducedWord.charAt(0);
		scanString = "";
		for (int j = 0; j < letterPosition[c1 - 'a']; ++j)
			scanString += ".";
		scanString += c1;		

		for (int j = 1; j < reducedWord.length(); ++j)
		{
			c1 = reducedWord.charAt(j - 1);
			c2 = reducedWord.charAt(j);
			int idx1 = c1 - 'a';
			int idx2 = c2 - 'a';
			int d = dist[idx1][idx2];
			for (int k = 0; k < d - 1; ++k)
				scanString += ".";
			scanString += c2;
		}
		c2 = reducedWord.charAt(reducedWord.length() - 1); // last letter in word
		int n = letterPosition[c2 - 'a'];
		for (int k = 0; k < keyboard.length - n - 1; ++k)
			scanString += ".";
		scanString += "S";
		String[] query = getCandidates(keystrokes, ew);
		if (query == null)
			return null;
		for (int k = 0; k < query.length; ++k)
			if (word.equals(query[k]))
				break;
			else
				scanString += ".";
		scanString += "W";
		return scanString;
	}

	//-------------------------------------------------------------------------------
	// OW = Optimized Word scanning method
	private static String getOWScanSequence(String wordArg, EncodedWord[] ewArg)
	{
		String word = wordArg;
		EncodedWord[] ew = ewArg;
		String scanString = "";
		int[] letterPosition = new int[26];
		String[] keyboard;
		letterPosition = getLetterPosition(ew);
		keyboard = getKeyboard(letterPosition);
		String keystrokes = getKeystrokes(word, letterPosition);		

		// build an array showing position of candidate word in list for
		// each word stem (e.g., "hello" --> h, he, hel, hell, hello)
		int[] candidateIndex = new int[word.length()];
		for (int j = 0; j < word.length(); ++j)
		{
			String keystrokeStem = keystrokes.substring(0, j + 1);
			String[] tryQuery = getCandidates(keystrokeStem, ew);
			if (tryQuery == null)
				return null;
			candidateIndex[j] = -1; // assume word not in candidate list
			for (int k = 0; k < tryQuery.length; ++k)
				if (word.equals(tryQuery[k]))
				{
					candidateIndex[j] = k;
					break;
				}
		}

		// use MLPK scan string as the starting point
		scanString = getMLPKScanSequence(word, ew);
		String optimizedScanString = "";
		String wordStem = "";
		for (int i = 0; i < scanString.length(); ++i)
		{
			char c = scanString.charAt(i);
			if (Character.isLowerCase(c))
			{
				String scanStem = scanString.substring(0, i + 1);
				wordStem += c;
				while (wordStem.length() < word.length())
				{
					char next = word.charAt(wordStem.length());
					if(letterPosition[c - 'a'] == letterPosition[next - 'a'])
						wordStem += next;
					else
						break;
				}
				int p1 = scanStem.length() + (keyboard.length - letterPosition[c - 'a']);
				int p2 = candidateIndex[wordStem.length() - 1];
				if (p2 < 0) p2 = 99999;
				if ((p1 + p2) < scanString.length())
				{
					optimizedScanString = scanStem;
					for (int j = 0; j < keyboard.length - letterPosition[c - 'a'] - 1; ++j)
						optimizedScanString += ".";
					optimizedScanString += "S";
					for (int j = 0; j < candidateIndex[wordStem.length() - 1]; ++j)
						optimizedScanString += ".";
					optimizedScanString += "W";
					break;
				}
			}
		}
		return optimizedScanString;
	}

	// -------------------------------------------------------------------
	// return the number of selections for scan string in a specified mode
	// -------------------------------------------------------------------
	public static int getSelectionCount(int modeArg, 
		String wordArg, String keystrokesArg, EncodedWord[] ewArg)
	{
		int mode = modeArg;
		int count = 0;
		String word = wordArg;
		String keystrokes = keystrokesArg;
		EncodedWord[] ew = ewArg;
		String scan = getScanSequence(mode, word, ew);

		// the following suffices for OLPS and MLPS methods
		if (mode == SAKStats.OLPS || mode == SAKStats.MLPS || 
			 mode == SAKStats.MLPK)
		{
			count = word.length() + 2; // +2 for final "S" and "W"
		}
		else if (mode == SAKStats.OW) // this one's tricky
		{
			word = scan.substring(0, scan.indexOf("S"));  // remove end (2 selections)
			String tmp = "";
			for (int i = 0; i < word.length(); ++i)
				if (word.charAt(i) != '.')
					tmp += word.charAt(i);
			for (int i = 0; i < keystrokes.length() - 1; ++i)
			{
				while (i < keystrokes.length() - 1 && keystrokes.charAt(i) == keystrokes.charAt(i + 1))
				{
					++count; // multiple selections, if possible
					++i; //test next pair
				}
			}
			count += tmp.length() + 2; // +2 for final "S" and "W"
		}
		return count;
	}

	// return the keystrokes for a word, given a letter-to-key mapping
	public static String getKeystrokes(String wordArg, int[] letterPositionArg)
	{
		String word = wordArg;
		int[] letterPosition = letterPositionArg;
		String keystrokes = "";
		for (int i = 0; i < word.length(); ++i)
			keystrokes += "" + (letterPosition[word.toLowerCase().charAt(i) - 'a'] + 1);
		return keystrokes;
	}

	// return a distance matrix give the scan distance between any two letters,
	// given a letter-to-key mapping
	private static int[][] getDistanceMatrix(int[] letterPositionArg)
	{
		int[] letterPosition = letterPositionArg;
		String[] keyboard = getKeyboard(letterPosition);
		int dist[][] = new int[26][26];
		for (int i = 0; i < 26; ++i)
		{
			for (int j = 0; j < 26; ++j)
			{	
				int p1 = letterPosition[i];
				int p2 = letterPosition[j];
				int d = p2 - p1;
				if (d > 0)
					dist[i][j] = d;
				else
					dist[i][j] = d + keyboard.length + 1;	
				//System.out.print(MyUtil.formatInt(dist[i][j], 2));
			}
			//System.out.println();
		}
		return dist;
	}

	// return a string array containing the letters on each key of virtual keyboard
	private static String[] getKeyboard(int[] letterPositionArg)
	{
		int[] letterPosition = letterPositionArg;
		String[] keyboard 
			= new String[letterPosition[25] + 1]; // 'z' position tells all
		for (int i = 0; i < keyboard.length; ++i)
			keyboard[i]= "";
		final String LETTERS = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
		for (int i = 0; i < 26; ++i)
			keyboard[letterPosition[i]] += LETTERS.charAt(i);
		return keyboard;
	}

	// return an int array containing the position of each letter on the keyboard
	public static int[] getLetterPosition(EncodedWord[] ewArg)
	{
		EncodedWord[] ew = ewArg;
		int[] letterPosition = new int[26];
		for (int i = 0; i < 26; ++i)
		{
			for (int j = 0; j < ew.length; ++j)
			{
				char c = (char)(i + 'a');
				int idx = ew[j].getWord().indexOf(c);
				if (idx >= 0)
				{
					letterPosition[i] 
						= Integer.parseInt(ew[j].getCode().substring(idx, idx + 1)) - 1;
					break;
				}				
			}
		}
		return letterPosition;
	}

	public static EncodedWord[] buildEncodedWordArray(String fileArg, 
		String[] keyboardArg) throws IOException
	{
		String file = fileArg;
		String[] keyboard = keyboardArg;	

		// open input file read word-freq data and build word-freq-keystroke data
		BufferedReader br = null;
      try
      {
         br = new BufferedReader(new FileReader(file));
      } catch (FileNotFoundException e)
      {
         System.out.println("File not found: " + file);
         System.exit(0);
      }

		Vector<String> w = new Vector<String>();
		Vector<String> f = new Vector<String>();
		String line;
		while ((line = br.readLine()) != null)
		{
			StringTokenizer st = new StringTokenizer(line);
			if (st.countTokens() != 2)
			{
				System.out.println("Format error in word-freq file!");
				System.exit(0);
			}
			w.addElement(st.nextToken());
			f.addElement(st.nextToken());
		}
		String[] wArray = new String[w.size()];
		String[] fArray = new String[f.size()];
		w.copyInto(wArray);
		f.copyInto(fArray);
		EncodedWord[] ew = new EncodedWord[wArray.length];
		for (int i = 0; i < ew.length; ++i)
		{
			String word = wArray[i].toLowerCase();
			int freq = Integer.parseInt(fArray[i]);
			String keystrokes = "";
			for (int j = 0; j < word.length(); ++j)
			{
				for (int k = 0; k < keyboard.length; ++k)
				{
					if (keyboard[k].indexOf(word.substring(j, j + 1)) >= 0)
					{
						keystrokes += k + 1;
						break;
					}
				}
			}
			ew[i] = new EncodedWord(word, freq, keystrokes);
		}
		Arrays.sort(ew, new EncodedWordByCode());
		return ew;
	}

	private static void usage()
	{
		System.out.println(
			"Usage: java SAKStats file letters [-olps|-mlps|-mlpk|-ow] [-wfk|-wfks|-spc|-sps]\n\n" +
			"where  file = word-freq dictionary file\n" +
			"       letters = letter-key assignment (e.g., abcdefg-hijklmn-opqrstu-vwxyz)\n" +
			"       -olps = one-letter-per-scan method\n" +
			"       -mlps = multiple-letter-per-scan method\n" +
			"       -mlpk = multiple-letter-per-key method\n" +
			"       -ow   = optimized-word method\n" +
			"       -wfk  = output word-freq-keystrokes data\n" +
			"       -wfks = output word-freq-keystrokes-scan data\n" +
			"       -spc  = output scans per character stat\n" + 
			"       -sps  = output selections per scan stat"
		);
		System.exit(0);
	}

	// return a query with candidate words
	private static String[] getCandidates(String keystrokesArg, EncodedWord[] ewArg)
	{
		String[] s = EncodedWord.getQuery(keystrokesArg, ewArg);
		if (s == null)
			return s;
		int size = s.length > 200 ? 200 : s.length;
		String[] s2 = new String[size];
		for (int i = 0; i < size; ++i)
			s2[i] = s[i];
		//System.out.print(" [s2.length=" + s2.length + "] ");
		//return s2;
		return s;
	}


	// return the number of possible MLPK opportunities for a word
	public static int getMLPKOpportunities(String wordArg, EncodedWord[] ewArg)
	{
		String word = wordArg;
		EncodedWord[] ew = ewArg;
		int[] letterPosition = getLetterPosition(ew);
		//String[] keyboard = getKeyboard(letterPosition);
		String keystrokes = getKeystrokes(word, letterPosition);
		//int dist[][] = getDistanceMatrix(letterPosition);
		//String scanString = "";

		boolean doubleUp = false;
		String reducedKeystrokes = "";
		reducedKeystrokes += keystrokes.charAt(0);
		//String reducedWord = "";
		//reducedWord += word.charAt(0);
		//char c1, c2;

		int opportunities = 0;
		for (int j = 1; j < keystrokes.length(); ++j)
		{
			char lastKeystroke = reducedKeystrokes.charAt(reducedKeystrokes.length() - 1);
			char c = keystrokes.charAt(j);
			if (lastKeystroke == c && !doubleUp)
			{
				doubleUp = true;
				++opportunities;
				continue;
			}
			else 
			{
				doubleUp = false;
				reducedKeystrokes += c;
				//reducedWord += word.charAt(j);
			}
		}
		return opportunities;
	}

	// return 1 if there is a OW opportunity, 0 otherwise
	public static int getOWOpportunities(String wordArg, EncodedWord[] ewArg)
	{
		int opportunities = 0;
		String word = wordArg;
		EncodedWord[] ew = ewArg;

		String scanString = "";
		int[] letterPosition = new int[26];
		String[] keyboard;
		letterPosition = getLetterPosition(ew);
		keyboard = getKeyboard(letterPosition);
		String keystrokes = getKeystrokes(word, letterPosition);		
		//String reducedKeystrokes = "" + keystrokes.charAt(0);
		//String reducedWord = "" + word.charAt(0);

		// build an array showing position of candidate word in list for
		// each word stem (e.g., "hello" --> h, he, hel, hell, hello)
		int[] candidateIndex = new int[word.length()];
		for (int j = 0; j < word.length(); ++j)
		{
			//System.out.println("j=" + j + ", word=" + word + ", keystrokes=" + keystrokes);
			String keystrokeStem = keystrokes.substring(0, j + 1);
			String[] tryQuery = getCandidates(keystrokeStem, ew);
			if (tryQuery == null)
				return 0;
			candidateIndex[j] = -1; // assume word not in candidate list
			for (int k = 0; k < tryQuery.length; ++k)
				if (word.equals(tryQuery[k]))
				{
					candidateIndex[j] = k;
					break;
				}
			//System.out.print(candidateIndex[j] + ", "); 
		}
		//System.out.println(); 
		//System.out.println(showWord);

		// use MLPK scan string as the starting point
		scanString = getMLPKScanSequence(word, ew);
		String optimizedScanString = "";
		String wordStem = "";
		for (int i = 0; i < scanString.length(); ++i)
		{
			//System.out.println("-------------");
			//System.out.println("i=" + i);
			char c = scanString.charAt(i);
			if (Character.isLowerCase(c))
			{
				String scanStem = scanString.substring(0, i + 1);
				wordStem += c;
				//System.out.println("word=" + word + ", wordStem=" + wordStem + ", scanStem=" + scanStem);
				// check if doubled-up letter (but not if already at end of word!)
				if (wordStem.length() < word.length())
				{
					char next = word.charAt(wordStem.length());
					if(letterPosition[c - 'a'] == letterPosition[next - 'a'])
						wordStem += next;
				}

				//String scanStem = scanString.substring(0, i + 1);
				int p1 = scanStem.length() + (keyboard.length - letterPosition[c - 'a']);
				//System.out.println("keyboard.length=" + keyboard.length + ", letterPosition[c]=" + letterPosition[c - 'a']);
				int p2 = candidateIndex[wordStem.length() - 1];
				if (p2 < 0) p2 = 99999;
				if ((p1 + p2) < scanString.length())
				{
					//System.out.println("Possible opportunity to optimize");
					//System.out.println("word=" + word + ", wordStem=" + wordStem + ", scanStem=" + scanStem);
					//System.out.println("candidateIndex= " + candidateIndex[wordStem.length() - 1]);
					optimizedScanString = scanStem;
					for (int j = 0; j < keyboard.length - letterPosition[c - 'a'] - 1; ++j)
						optimizedScanString += ".";
					optimizedScanString += "S";
					for (int j = 0; j < candidateIndex[wordStem.length() - 1]; ++j)
						optimizedScanString += ".";
					optimizedScanString += "W";
					//System.out.println("optimizedScanString=" + optimizedScanString);	
					//if (optimizedScanString.length() > scanString.length())
					//	System.out.println("Oops!!!");

					// the moment of truth...
					if (optimizedScanString.length() < scanString.length())
						opportunities = 1;
					break;
				}
			}
		}
		return opportunities;
	}
}